from warnings import warn

import torch
import itertools
from typing import Iterable
from .user_feature import UserFeature
from .metrics import calculate, MetricEnum


def __mean(v):
    return torch.mean(v).item() if isinstance(v, torch.Tensor) else v


def _calculate_for_feature(
    group: UserFeature,
    metrics: Iterable[MetricEnum],
    logits: torch.Tensor,
    targets: torch.Tensor = None,
    k: int = 10,
    return_individual=False,
):
    """
    Computes the values for a given list of metrics for the users with different demographic features.
    Moreover, pairwise differences between the group metrics are also calculated.

    In the context of gender, the metrics would be computed for male users and female users individually.

    :param group: A user feature for which to compute the metrics.
    :param metrics: The list of metrics to compute. Check out 'supported_metrics' for a list of names.
    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param return_individual: Whether the results for individual users should also be returned.
    :return: A dictionary of metrics computed for the individual user groups, and their pairwise differences
     in the form of {metric_name: value} pairs.
    """
    warn(
        "This method is deprecated. Please use `rmet.groups.calculate_per_group` instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    results = {}

    # calculate metrics for users of a single feature
    for lbl, indices in group:
        t = targets[indices] if targets is not None else None
        results[f"{group.name}_{lbl}"] = calculate(
            metrics, logits[indices], t, k, return_per_user=return_individual
        )

    pairs = list(itertools.combinations(group.unique_labels, 2))
    for a, b in pairs:
        pair_results = dict()
        for m in metrics:
            pair_results[str(m)] = (
                results[f"{group.name}_{a}"][m] - results[f"{group.name}_{b}"][m]
            )
        results[f"{group.name}_{a}-{b}"] = pair_results
    return results


def _dict_difference(first: any, second: any):
    if not isinstance(first, type(second)):
        raise ValueError(f"Types mismatch! ({type(first)=} vs {type(second)=})")

    if isinstance(first, dict):
        if set(first.keys()) != set(second.keys()):
            raise ValueError("Dictionaries do not have the same keys!")
        return {k: _dict_difference(first[k], second[k]) for k in first.keys()}

    else:
        return first - second


def calculate_for_feature(
    group: UserFeature,
    metrics: Iterable[MetricEnum | str],
    logits: torch.Tensor,
    targets: torch.Tensor = None,
    k: int | Iterable[int] = 10,
    return_individual: bool = False,
    flatten_results: bool = False,
    flattened_parts_separator: str = "/",
    flattened_results_prefix: str = "",
):
    """
    Computes the values for a given list of metrics for the users with different demographic features.
    Moreover, pairwise differences between the group metrics are also calculated.

    In the context of gender, the metrics would be computed for male users and female users individually.

    :param group: A user feature for which to compute the metrics.
    :param metrics: The list of metrics to compute. Check out 'supported_metrics' for a list of names.
    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param return_individual: Whether the results for individual users should also be returned.
    :param flatten_results: Whether to flatten the results' dictionary.
                            Key is of format "{prefix}/{metric}@{k}/{feature_group}" for separator "/"
    :param flattened_parts_separator: How to separate the individual parts of the flattened key
    :param flattened_results_prefix: Prefix to prepend to the flattened results key.
    :return: A dictionary of metrics computed for the individual user groups, and their pairwise differences
     in the form of {metric_name: value} pairs.
    """
    warn(
        "This method is deprecated. Please use `rmet.groups.calculate_per_group` instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    results = dict()

    # calculate metrics for users of a single feature
    for lbl, indices in group:
        t = targets[indices] if targets is not None else None
        results[f"{group.name}_{lbl}"] = calculate(
            metrics, logits[indices], t, k, return_per_user=return_individual
        )

    # calculate the differences between features
    pairs = list(itertools.combinations(group.unique_labels, 2))
    for a, b in pairs:
        # only iterate over metrics, and not all values in results, as it might contain
        # results on the user-level as well
        results[f"{group.name}_{a}-{b}"] = {
            str(m): _dict_difference(
                results[f"{group.name}_{a}"][m], results[f"{group.name}_{b}"][m]
            )
            for m in metrics
        }

    if flatten_results:
        flattened_results = dict()
        for group, group_results in results.items():
            for metric, metric_results in group_results.items():
                for k, results_for_k in metric_results.items():
                    # collect parts out of which final key should be generated
                    key_parts = (
                        [flattened_results_prefix]
                        if flattened_results_prefix
                        else list()
                    )
                    key_parts.extend([f"{metric}@{k}", group])

                    key = flattened_parts_separator.join(key_parts)
                    flattened_results[key] = results_for_k
        return flattened_results

    return results



import math
import torch
import unittest
import numpy as np
import scipy.sparse as sp

from rmet.metrics import (
    calculate,
    supported_metrics,
)

from rmet.groups import calculate_per_group
from rmet.batch_evaluator import BatchEvaluator, EvaluatorResults


class TestRecommenderGroupMetrics(unittest.TestCase):

    def setUp(self):
        # top-k: [0, 1, 3], [3, 4, 2]
        self.logits = [[0.9, 0.8, 0, 0.4, 0], [0.0, 0.0, 0.3, 0.9, 0.8]]
        # relevant: [0, 2, 3], [3, 4]
        self.targets = [[1, 0, 1, 1, 0], [0, 0, 0, 1, 1]]

        self.supported_types = {
            "torch": {
                "cast": torch.tensor,
                "cast-result": torch.tensor,
                "all_close": torch.allclose,
                "zeros_like": torch.zeros_like,
            },
            "np": {
                "cast": np.array,
                "cast-result": np.array,
                "all_close": np.allclose,
                "zeros_like": np.zeros_like,
            },
            "sparse-csr": {
                "cast": sp.csr_array,
                "cast-result": np.array,
                "all_close": np.allclose,
                "zeros_like": np.zeros_like,
            },
        }

        self.k = [3, 5]

        rng = np.random.default_rng(42)
        self.n_users, self.n_items = 5, 10
        self.logits = rng.random(size=(self.n_users, self.n_items))
        self.targets = rng.integers(low=0, high=2, size=(self.n_users, self.n_items))
        self.item_ranks = np.argsort(self.targets.sum(0))

        self.group_assignments = {
            "G1": rng.choice(["A", "B", "C"], size=self.n_users),
            "G2": rng.choice(["X", "Y", "Z"], size=self.n_users),
        }

        self.metrics = supported_metrics

    def _extract_at_level(
        self, d: dict, depth: int, key: any, keep_level: bool = False
    ):
        """
        Extracts the dictionary values of a given depth that match a certain key
        of a nested dictionary
        """
        if isinstance(d, dict):
            new_dict = {}
            for k, v in d.items():
                if depth == 0:
                    if k == key:
                        if keep_level:
                            new_dict[k] = v
                        else:
                            return v
                elif depth > 0:
                    new_dict[k] = self._extract_at_level(
                        v, depth - 1, key, keep_level=keep_level
                    )
            return new_dict
        return d

    def _extract_at_levels(
        self, d: dict, depth_keys: dict[int, any], keep_levels: bool = False
    ):
        """
        Extracts the values of a nested dictionary based on the given depths and corresponding keys.
        """
        # sort by depths so that we go from most shallow to deepest depths for efficiency
        sorted_depths = sorted(depth_keys.items(), key=lambda x: x[0], reverse=False)

        for i, (depth, key) in enumerate(sorted_depths):
            if not keep_levels:
                # -i to account for dropped depths
                depth -= i
            d = self._extract_at_level(d, depth, key, keep_level=keep_levels)

        return d

    def _assert_dictionary_equality(self, d1, d2, is_close_fn):
        # check matching keys
        self.assertEqual(type(d1), type(d2))
        self.assertSetEqual(set(d1.keys()), set(d2.keys()))
        for k in d1:
            # dict might be nested
            if isinstance(d1[k], dict):
                self._assert_dictionary_equality(d1[k], d2[k], is_close_fn)
            else:
                # check matching results
                self.assertEqual(type(d1[k]), type(d2[k]))
                if isinstance(d1[k], (np.ndarray, torch.Tensor)):
                    self.assertTrue(is_close_fn(d1[k], d2[k]))
                else:
                    # ignore values that are NaN (equality comparison doesn't work for them)
                    if math.isnan(d1[k]) and math.isnan(d2[k]):
                        pass
                    else:
                        self.assertAlmostEqual(d1[k], d2[k], delta=1e-4)

    def test_compute_nested(self):
        for _, fn_lookup in self.supported_types.items():
            for group_name, group_assignment in self.group_assignments.items():
                results = calculate_per_group(
                    metrics=self.metrics,
                    group_name=group_name,
                    group_assignment=group_assignment,
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                    return_aggregated=True,
                    return_per_user=True,
                    calculate_std=True,
                    flatten_results=False,
                )

                # test results
                for group in np.unique(group_assignment):
                    mask = group_assignment == group

                    # we can use this function as it's already tested elsewhere
                    expected = calculate(
                        metrics=self.metrics,
                        logits=fn_lookup["cast"](self.logits)[mask],
                        targets=fn_lookup["cast"](self.targets)[mask],
                        k=self.k,
                        return_aggregated=True,
                        return_per_user=True,
                        calculate_std=True,
                        flatten_results=False,
                    )

                    self._assert_dictionary_equality(
                        self._extract_at_levels(results, {2: group_name, 3: group}),
                        expected,
                        fn_lookup["all_close"],
                    )

    def test_compute_flattened(self):
        for _, fn_lookup in self.supported_types.items():
            for group_name, group_assignment in self.group_assignments.items():
                results = calculate_per_group(
                    metrics=self.metrics,
                    group_name=group_name,
                    group_assignment=group_assignment,
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                    return_aggregated=True,
                    return_per_user=True,
                    calculate_std=True,
                    flatten_results=True,
                )

                # test results
                for group in np.unique(group_assignment):
                    mask = group_assignment == group
                    group_result_suffix = f"/{group_name}_{group}"

                    # we can use this function as it's already tested elsewhere
                    expected = calculate(
                        metrics=self.metrics,
                        logits=fn_lookup["cast"](self.logits)[mask],
                        targets=fn_lookup["cast"](self.targets)[mask],
                        k=self.k,
                        return_aggregated=True,
                        return_per_user=True,
                        calculate_std=True,
                        flatten_results=True,
                        flatten_suffix=group_result_suffix,
                    )

                    self._assert_dictionary_equality(
                        {
                            k: v
                            for k, v in results.items()
                            if k.endswith(group_result_suffix)
                        },
                        expected,
                        fn_lookup["all_close"],
                    )

    def test_compute_with_best_logits_input(self):
        for _, fn_lookup in self.supported_types.items():
            for group_name, group_assignment in self.group_assignments.items():
                _, best_logits_indices = calculate(
                    metrics=self.metrics,
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                    return_best_logit_indices=True,
                )

                results = calculate_per_group(
                    metrics=self.metrics,
                    group_name=group_name,
                    group_assignment=group_assignment,
                    best_logit_indices=fn_lookup["cast-result"](best_logits_indices),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                    return_aggregated=True,
                    return_per_user=True,
                    calculate_std=True,
                    flatten_results=True,
                    n_items=self.n_items,
                )

                # test results
                for group in np.unique(group_assignment):
                    mask = group_assignment == group
                    group_result_suffix = f"/{group_name}_{group}"

                    # we can use this function as it's already tested elsewhere
                    expected = calculate(
                        metrics=self.metrics,
                        best_logit_indices=fn_lookup["cast-result"](
                            best_logits_indices
                        )[mask],
                        targets=fn_lookup["cast"](self.targets)[mask],
                        k=self.k,
                        return_aggregated=True,
                        return_per_user=True,
                        calculate_std=True,
                        flatten_results=True,
                        flatten_suffix=group_result_suffix,
                        n_items=self.n_items,
                    )

                    self._assert_dictionary_equality(
                        {
                            k: v
                            for k, v in results.items()
                            if k.endswith(group_result_suffix)
                        },
                        expected,
                        fn_lookup["all_close"],
                    )

    def test_group_batch_evaluator(self):
        for _, fn_lookup in self.supported_types.items():
            batch_evaluator = BatchEvaluator(
                metrics=self.metrics,
                top_k=self.k,
                calculate_std=True,
                n_items=None,
                item_ranks=fn_lookup["cast-result"](self.item_ranks),
            )

            for i in range(len(self.logits)):
                batch_evaluator.eval_batch(
                    user_indices=fn_lookup["cast-result"]([i]),
                    # slice to maintain batch
                    logits=fn_lookup["cast"](self.logits[i : i + 1]),
                    targets=fn_lookup["cast"](self.targets[i : i + 1]),
                    group_assignments={
                        k: v[i : i + 1] for k, v in self.group_assignments.items()
                    },
                )
            result = batch_evaluator.get_results()

            expected = EvaluatorResults(
                user_level_metrics={},
                aggregated_metrics={},
                user_indices=None,
                user_top_k=None,
            )

            # compute expected results
            overall_results = calculate(
                metrics=self.metrics,
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
                calculate_std=True,
                return_per_user=True,
                return_aggregated=True,
                flatten_results=True,
                item_ranks=fn_lookup["cast-result"](self.item_ranks),
            )
            expected.aggregated_metrics.update(
                {k: v for k, v in overall_results.items() if "_user" not in k}
            )
            expected.user_level_metrics.update(
                {
                    k.replace("_user", ""): v
                    for k, v in overall_results.items()
                    if "_user" in k
                }
            )

            for group_name, group_assignment in self.group_assignments.items():
                for group in np.unique(group_assignment):
                    mask = group_assignment == group
                    group_result_suffix = f"/{group_name}_{group}"

                    # we can use this function as it's already tested elsewhere
                    expected_per_group_value = calculate(
                        metrics=self.metrics,
                        logits=fn_lookup["cast-result"](self.logits)[mask],
                        targets=fn_lookup["cast"](self.targets)[mask],
                        k=self.k,
                        return_aggregated=True,
                        return_per_user=True,
                        calculate_std=True,
                        flatten_results=True,
                        flatten_suffix=group_result_suffix,
                        n_items=self.n_items,
                        item_ranks=fn_lookup["cast-result"](self.item_ranks),
                    )

                    expected.aggregated_metrics.update(
                        {
                            k: v
                            for k, v in expected_per_group_value.items()
                            if "_user" not in k
                        }
                    )
                    expected.user_level_metrics.update(
                        {
                            k.replace("_user", ""): v
                            for k, v in expected_per_group_value.items()
                            if "_user" in k
                        }
                    )

            self._assert_dictionary_equality(
                result.user_level_metrics,
                expected.user_level_metrics,
                fn_lookup["all_close"],
            )
            self._assert_dictionary_equality(
                result.aggregated_metrics,
                expected.aggregated_metrics,
                fn_lookup["all_close"],
            )


if __name__ == "__main__":
    unittest.main()



import torch
import numpy as np
import scipy.sparse as sp
from typing import Iterable
from natsort import natsorted
from dataclasses import dataclass
from collections import defaultdict

from .metrics import (
    calculate,
    supported_metrics,
    supported_user_accuracy_metrics,
    supported_global_metrics,
    supported_user_beyond_accuracy_metrics,
)

from .groups import calculate_per_group
from .type_helpers import std


@dataclass
class EvaluatorResults:
    aggregated_metrics: dict[str, torch.Tensor]
    user_level_metrics: dict[str, torch.Tensor]
    user_indices: torch.Tensor
    user_top_k: torch.Tensor


class BatchEvaluator:
    """
    Helper class that supports evaluating recommendations batch after batch, stores these
    internally and can aggregate the results after the final batch.
    """

    def __init__(
        self,
        metrics: Iterable[str],
        top_k: Iterable[int],
        calculate_std: bool = True,
        n_items: int = None,
        metric_prefix: str = "",
        **kwargs,
    ):
        """
        Initializes a BatchEvaluator instance based on the provided config.

        :param metrics:         Metric name or list of metrics to compute. Check out 'supported_metrics'
                                for a list of all available metrics.
        :param k:               Top k items to consider
        :param calculate_std:   Whether to calculate the standard deviation for the aggregated results
        :param n_items:         Number of items in dataset, in case targets do not contain 'labels' for all items
        :param metric_prefix:   Prefix for the computed metrics
        :param kwargs:          Additional parameters that are passed to metric computations, e.g., item_ranks
        """
        self.metrics = metrics
        self.top_k = top_k
        self.calculate_std = calculate_std

        self.n_items = n_items
        self.metric_prefix = metric_prefix
        self.calculation_kwargs = kwargs

        # ensure that only valid metrics are supplied
        invalid_metrics = set(self.metrics) - set(supported_metrics)
        if len(invalid_metrics) > 0:
            raise ValueError(
                f"Metric(s) {invalid_metrics} are not supported. "
                f"Select metrics from {supported_metrics}."
            )

        # determine to which kind of metrics the different metrics belong to
        # need to know whether we can compute the metrics per batch, or only
        # after the final batch has been processed
        # Note: We don't want to compute everything at the end, as we might not be able
        #       to make use GPU computations like that.
        self._user_metrics = set(self.metrics).intersection(
            set(supported_user_accuracy_metrics).union(
                supported_user_beyond_accuracy_metrics
            )
        )
        self._dist_metrics = set(self.metrics).intersection(supported_global_metrics)

        self._are_results_available = False
        self._user_indices = None
        self._user_top_k = None
        self._user_group_assignments = None

        # internal storage for the results
        self._user_level_results = None
        self._user_group_level_results = None
        self._reset_internal_dict()

    def _reset_internal_dict(self):
        """
        Resets the internal memory on computed and gathered metrics.
        """
        self._user_level_results = defaultdict(lambda: list())
        self._user_group_level_results = defaultdict(lambda: list())
        self._user_indices = list()
        self._user_top_k = list()
        self._user_group_assignments = defaultdict(lambda: list())
        self._are_results_available = False

    def _detach_and_to_cpu(self, a: any):
        if isinstance(a, dict):
            return {k: self._detach_and_to_cpu(v) for k, v in a.items()}
        elif isinstance(a, torch.Tensor):
            return a.detach().cpu()
        else:
            return a

    def _concat_list_elements(self, a: tuple | list):
        if len(a) > 0:
            if isinstance(a[0], torch.Tensor):
                return torch.cat(a)
            elif isinstance(a[0], np.ndarray):
                return np.concatenate(a)
        return a

    def _concat_dict_list_values(self, d: dict):
        return {k: self._concat_list_elements(v) for k, v in d.items()}

    @staticmethod
    def _natsort_dict(d: dict):
        return {k: d[k] for k in natsorted(d.keys())}

    def _calculate_user_metrics(self, logits: torch.Tensor, y_true: torch.Tensor):
        """
        Wrapper function to compute user-based metrics, e.g., recall and precision
        """
        user_level_results, top_k_indices = calculate(
            metrics=self._user_metrics,
            logits=logits,
            targets=y_true,
            k=self.top_k,
            return_aggregated=False,
            return_per_user=True,
            flatten_results=True,
            flatten_prefix=self.metric_prefix,
            n_items=self.n_items,
            return_best_logit_indices=True,
            **self.calculation_kwargs,
        )

        # drop the "_user" part as it's not relevant for us
        user_level_results = {
            k.replace("_user", ""): v
            for k, v in user_level_results.items()
            if k.endswith("_user")
        }
        return user_level_results, top_k_indices

    def _calculate_user_group_metrics(
        self,
        top_k_indices: torch.Tensor,
        y_true: torch.Tensor,
        group_assignments: dict[str, Iterable],
    ):
        """
        Wrapper function to compute user-based group metrics, e.g., recall and precision per group
        """
        results = {}
        for group_name, group_assignment in group_assignments.items():
            user_level_group_results = calculate_per_group(
                group_name=group_name,
                group_assignment=group_assignment,
                metrics=self._user_metrics,
                targets=y_true,
                k=self.top_k,
                return_aggregated=False,
                return_per_user=True,
                flatten_results=True,
                flatten_prefix=self.metric_prefix,
                n_items=self.n_items,
                best_logit_indices=top_k_indices,
                **self.calculation_kwargs,
            )

            # drop the "_user" part as it's not relevant for us
            for k, v in user_level_group_results.items():
                results[k.replace("_user", "")] = v
        return results

    def _calculate_distribution_metrics(self, user_top_k):
        """
        Wrapper function to compute distribution-based metrics, e.g., coverage.
        """
        results = calculate(
            metrics=self._dist_metrics,
            k=self.top_k,
            return_aggregated=True,
            return_per_user=False,
            flatten_results=True,
            flatten_prefix=self.metric_prefix,
            best_logit_indices=user_top_k,
            n_items=self.n_items,
            **self.calculation_kwargs,
        )
        return results

    def _calculate_group_distribution_metrics(
        self, user_top_k, group_assignments: dict[str, Iterable]
    ):
        """
        Wrapper function to compute group-based distribution metrics, e.g., coverage per group.
        """
        results = {}
        for group_name, group_assignment in group_assignments.items():
            group_results = calculate_per_group(
                group_name=group_name,
                group_assignment=group_assignment,
                metrics=self._dist_metrics,
                k=self.top_k,
                return_aggregated=True,
                return_per_user=False,
                flatten_results=True,
                flatten_prefix=self.metric_prefix,
                best_logit_indices=user_top_k,
                n_items=self.n_items,
                **self.calculation_kwargs,
            )
            results.update(group_results)
        return results

    @torch.no_grad()
    def eval_batch(
        self,
        user_indices: torch.Tensor | np.ndarray,
        logits: torch.Tensor | np.ndarray | sp.csr_array,
        targets: torch.Tensor | np.ndarray | sp.csr_array,
        group_assignments: dict[str, Iterable] = None,
    ):
        """
        Evaluates a batch of logits and their true targets and stores the results internally.
        To retrieve the results, call `get_results`.

        :param user_indices:    indices of users in batch
        :param logits:          predicted ratings, expected shape is (batch_size, n_items)
        :param targets:         target ratings, expected  (batch_size, n_items)
        :param group_assignments: if given, metrics will also be computed on group-level
        """
        if logits.shape != targets.shape:
            raise ValueError(
                f"Logits and true labels must have the same shape ({logits.shape} != {targets.shape})"
            )
        if logits.ndim != 2:
            raise ValueError(
                f"Logits and targets are expected to have 2 dimensions "
                f"instead of {logits.ndim}."
            )

        if self.n_items is None:
            self.n_items = targets.shape[-1]

        # compute and store user-based metrics
        user_metrics, top_k_indices = self._calculate_user_metrics(logits, targets)
        user_metrics = self._detach_and_to_cpu(user_metrics)
        for metric_name, metric_values in user_metrics.items():
            self._user_level_results[metric_name].append(metric_values)

        # compute and store user group-level metrics
        if group_assignments is not None:
            user_group_metrics = self._calculate_user_group_metrics(
                top_k_indices, targets, group_assignments
            )
            user_group_metrics = self._detach_and_to_cpu(user_group_metrics)
            for metric_name, metric_values in user_group_metrics.items():
                self._user_group_level_results[metric_name].append(metric_values)

            for g, assignments in group_assignments.items():
                # for simplicity we assume assignmets are convertable to numpy arrays
                self._user_group_assignments[g].append(np.array(assignments))

        user_indices = self._detach_and_to_cpu(user_indices)
        top_k_indices = self._detach_and_to_cpu(top_k_indices)

        self._user_indices.append(user_indices)
        self._user_top_k.append(top_k_indices)

        self._are_results_available = True

    def get_results(self, reset_state: bool = True) -> EvaluatorResults:
        """
        Retrieves all user- and distribution-based results.

        :param reset_state: Whether to reset internal memory. If true, `get_results` can only
                            be called once.
        """
        if not self._are_results_available:
            raise RuntimeError(
                "No results have yet been calculated. Call `eval_batch` before "
                "calling this method."
            )

        user_top_k = self._concat_list_elements(self._user_top_k)
        user_indices = self._concat_list_elements(self._user_indices)

        aggregated_results, user_level_results = {}, {}
        if len(self._user_metrics) > 0:
            # join user metrics across all batches
            all_user_results = self._user_level_results | self._user_group_level_results
            user_level_results = self._concat_dict_list_values(all_user_results)

            # aggregate the results
            aggregated_results = {
                k: v.mean().item() for k, v in user_level_results.items()
            }

            if self.calculate_std:
                # determine all flattened metrics based on user-level
                # (group-level results have groups appended)
                # e.g., ndcg@10 -> ndcg@10_std
                user_metrics = list(self._user_level_results.keys())
                # first try replace longest, then shortest, so that
                # we don't replace ndcg@100 with the value for ndcg@10
                user_metrics = sorted(user_metrics, key=len, reverse=True)
                std_replacement_strings = {m: m + "_std" for m in user_metrics}

                for k, v in user_level_results.items():
                    for original, replacement in std_replacement_strings.items():
                        if original in k:
                            # limit replacement to a single time
                            k = k.replace(original, replacement)
                            break
                    aggregated_results[k] = std(v)

        # calculate distribution metrics
        if len(self._dist_metrics) > 0:
            distribution_results = self._calculate_distribution_metrics(user_top_k)
            aggregated_results.update(distribution_results)

            # calculate group-level distribution metrics
            if len(self._user_group_assignments) > 0:
                # first join all the gathered groups
                group_assignments = self._concat_dict_list_values(
                    self._user_group_assignments
                )
                group_distribution_results = self._calculate_group_distribution_metrics(
                    user_top_k, group_assignments
                )
                aggregated_results.update(group_distribution_results)

        if reset_state:
            self._reset_internal_dict()

        return EvaluatorResults(
            # employ natural sorting on metrics, so that, e.g., @20 comes before @100
            aggregated_metrics=self._natsort_dict(aggregated_results),
            user_level_metrics=self._natsort_dict(user_level_results),
            user_indices=user_indices,
            user_top_k=user_top_k,
        )



import torch
import numpy as np
import scipy.sparse as sp
from typing import Iterable


def assert_supported_type(a: any):
    if not isinstance(a, (np.ndarray, torch.Tensor, sp.csr_array)):
        raise TypeError(f"Type {type(a)} of input not supported.")


def zeros_like_float(a: torch.Tensor | np.ndarray):
    if isinstance(a, torch.Tensor):
        return torch.zeros_like(a, dtype=torch.float, device=a.device)

    elif isinstance(a, np.ndarray):
        return np.zeros_like(a, dtype=float)

    else:
        assert_supported_type(a)


def zeros_float(shape, reference_a):
    if isinstance(reference_a, torch.Tensor):
        return torch.zeros(shape, device=reference_a.device)

    elif isinstance(reference_a, np.ndarray | sp.csr_array):
        return np.zeros(shape)
    else:
        assert_supported_type(reference_a)


def as_float(a: torch.Tensor | np.ndarray):
    if isinstance(a, torch.Tensor):
        return a.float()

    elif isinstance(a, np.ndarray):
        return a.astype(float)

    else:
        assert_supported_type(a)


def stack(a: Iterable[np.ndarray] | Iterable[torch.Tensor]):
    if len(a) == 0:
        raise ValueError("Cannot perform stack on empty iterable.")

    if isinstance(a[0], torch.Tensor):
        return torch.stack(a)
    else:
        return np.stack(a)


def generate_non_zero_mask(a: torch.Tensor | np.ndarray, dim=-1):
    non_zero_counter = a.sum(-1)

    if isinstance(non_zero_counter, torch.Tensor):
        return torch.argwhere(non_zero_counter).flatten()

    elif isinstance(non_zero_counter, np.ndarray | sp.csr_array):
        return np.argwhere(non_zero_counter).flatten()

    else:
        assert_supported_type(non_zero_counter)


def get_unique_values(top_indices: torch.Tensor | np.ndarray, sorted=False):
    if isinstance(top_indices, torch.Tensor):
        return top_indices.unique(sorted=sorted)

    elif isinstance(top_indices, np.ndarray):
        return np.unique(top_indices)

    else:
        assert_supported_type(top_indices)


def std(a: torch.Tensor | np.ndarray):
    if isinstance(a, torch.Tensor):
        return torch.std(a).item()

    elif isinstance(a, np.ndarray):
        # set degrees of freedom to 1 to have same results as torch
        return np.std(a, ddof=1).item()

    else:
        assert_supported_type(top_indices)


def _get_top_k_numpy(a: np.ndarray, k: int, sorted: bool = True):
    # use partition which is much faster than argsort() on big arrays
    indices_unsorted = np.argpartition(a, -k, axis=-1)[:, -k:]
    if not sorted:
        return indices_unsorted

    # sort indices by their values
    values_unsorted = np.take_along_axis(a, indices_unsorted, axis=-1)
    sorting = np.argsort(values_unsorted, axis=-1)
    indices_sorted = np.take_along_axis(indices_unsorted, sorting, axis=-1)

    # reverse order from high to low
    return indices_sorted[:, ::-1]


def _get_top_k_sparse(a: sp.csr_array, k: int):
    n_rows, n_cols = a.shape

    if k > n_cols:
        raise ValueError("k must be at most number of columns")

    # create placeholder to fill
    top_k_indices = np.zeros(shape=(n_rows, k), dtype=int)

    # go row by row over array
    for i in range(n_rows):
        row = a[i, :]
        data = row.data

        if isinstance(row, sp.csr_array):
            ind = row.indices
        elif isinstance(row, sp.coo_array):
            ind = row.coords[0]
        else:
            raise TypeError(f"Type {type(row)} for rows in array not expected.")

        # we stay aligned with behaviour of np.argsort, which orders
        # duplicate values by their indices of occurence (first come first)
        # note that we need only to consider at most k indices as padding candidates,
        candidate_padding_indices = np.arange(n_cols - k, n_cols)

        # select those candidates that don't already occur
        mask = np.isin(candidate_padding_indices, ind, invert=True)
        padding_indices = candidate_padding_indices[mask]
        full_indices = np.concatenate([padding_indices, ind])

        # we consider all padding indices to have data=0, which is the default behaviour
        # for scipy. This way, negative data values are lower ranked than the padded values
        padding_data = np.zeros_like(padding_indices, dtype=a.dtype)

        # sorting is always required as there might be negative values in original data
        full_data = np.concatenate([padding_data, data])
        sorting = np.argsort(full_data)
        full_indices = full_indices[sorting]

        # select final indices
        top_k_row = full_indices[-k:][::-1]
        top_k_indices[i] = top_k_row

    return top_k_indices


def get_top_k(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    k=10,
    logits_are_top_indices: bool = False,
    sorted: bool = True,
):
    """
    Gets the top-k indices for the logits

    :param logits: prediction matrix about item relevance
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    :param sorted: whether indices should be returned in sorted order
    """
    if logits_are_top_indices:
        return logits[:, :k]

    else:
        if isinstance(logits, torch.Tensor):
            return logits.topk(k, dim=-1, sorted=sorted).indices

        elif isinstance(logits, np.ndarray):
            return _get_top_k_numpy(logits, k, sorted=sorted)

        elif isinstance(logits, sp.csr_array):
            return _get_top_k_sparse(logits, k)

        else:
            assert_supported_type(logits)



import torch
import numpy as np
import scipy.sparse as sp
from collections import defaultdict
from .metrics import calculate
from .type_helpers import get_unique_values


def _try_masking(mask, a: any = None):
    """
    Helper function to try and mask a given array, otherwise
    return None.
    """
    if a is None:
        return None
    return a[mask]


def _nested_default_to_normal_dict(d: defaultdict):
    if isinstance(d, (dict, defaultdict)):
        return {k: _nested_default_to_normal_dict(v) for k, v in d.items()}
    return d


def _create_nested_defaultdict(depth: int):
    if depth < 1:
        raise ValueError("Depth must be at least 1")
    if depth == 1:
        return defaultdict()
    return defaultdict(lambda: _create_nested_defaultdict(depth - 1))


def calculate_per_group(
    group_name: str,
    group_assignment: any,
    logits: torch.Tensor | np.ndarray | sp.csr_array = None,
    targets: torch.Tensor | np.ndarray | sp.csr_array = None,
    flatten_results: bool = False,
    flatten_prefix: str = "",
    best_logit_indices: torch.Tensor | np.ndarray = None,
    return_best_logit_indices: bool = False,
    **kwargs,
):
    """
    :param group_name:          for labeling the group
    :param group_assignment:    assignment which user belongs to which group to compute
                                group-based metrics
    :param logits:              prediction matrix about item relevance
    :param targets:             0/1 matrix encoding true item relevance, same shape as logits
    :param flatten_results:     whether to flatten the results' dictionary.
    :param flatten_prefix:      prefix to use for flattened results, e.g., to differentiate between
                                different splits
    :param best_logit_indices:  previously computed indices of the best logits in sorted order
    :param return_best_logit_indices:   whether to return the indices of the best logits
    :param **kwargs:            see documentation of `rmet.metrics.calculate()`

    :return: a nested dictionary depending of structure {
            `metric`: {
                `k`: {
                    `key`: {
                        `group_name`: {
                            `group_value`: `value`
                        }
                    }
                }
            },
             where `key` depends on the metric and parameters:
             - "user":   ..., # for user-based metrics if 'return_per_user'
             - "mean":   ..., # for user-based metrics if 'return_aggregated'
             - "std":    ..., # for user-based metrics if 'return_aggregated' and 'calculate_std'
             - "global": ..., # for global metrics
             or a nested dictionary with keys in the format
             - for "mean" and "global" keys due to their importance:
               `{flatten_prefix}/{metric}@{k}/{group_name}_{group_value}`
             - for "std" and "user" keys
               `{flatten_prefix}/{metric}@{k}_{key}/{group_name}_{group_value}`
    """

    # we'd like to process groups as either torch tensors or
    # numpy arrays for simpler handling
    if not isinstance(group_assignment, torch.Tensor):
        group_assignment = np.array(group_assignment)
    # unique_values = get_unique_values(group_assignment)
    unique_values = set(list(group_assignment))

    # iterate over all groups and collect results
    # metric -> k -> group_name -> group_value -> result
    results_per_group = _create_nested_defaultdict(4)
    # group_name -> group_value -> logits
    logits_per_group = _create_nested_defaultdict(2)
    for v in unique_values:
        mask_matching_samples = group_assignment == v

        # compute results for given group
        group_results = calculate(
            logits=_try_masking(mask_matching_samples, logits),
            targets=_try_masking(mask_matching_samples, targets),
            best_logit_indices=_try_masking(mask_matching_samples, best_logit_indices),
            flatten_results=flatten_results,
            # we append prefix before the group results if flattened, not per group
            flatten_prefix=flatten_prefix,
            flatten_suffix=f"/{group_name}_{v}",
            return_best_logit_indices=return_best_logit_indices,
            **kwargs,
        )

        # split results based on supplied arguments
        if return_best_logit_indices:
            metrics_results, best_logits = group_results
        else:
            metrics_results, best_logits = group_results, None

        # store results
        if flatten_results:
            # for simplicity we do the flattening in calculate()
            results_per_group.update(metrics_results)
        else:
            for m, m_results in metrics_results.items():
                for ki, ki_results in m_results.items():
                    results_per_group[m][ki][group_name][v] = ki_results
        logits_per_group[group_name][v] = best_logits

    results_per_group = _nested_default_to_normal_dict(results_per_group)
    logits_per_group = _nested_default_to_normal_dict(logits_per_group)

    if return_best_logit_indices:
        return results_per_group, logits_per_group
    return results_per_group



import numpy as np
from warnings import warn


class UserFeature:
    def __init__(self, name: str, labels: list):
        """
        Splits users based on some arbitrary feature in different groups. This is used
        to ease calculating differences of recommendation systems for users with different demographics, e.g., gender.

        :param name: The name of the feature.
        :param labels: The labels for the individual users of the feature. The users are grouped based on them.
        """
        warn(
            f"{self.__class__.__name__} is deprecated. Please use `rmet.groups.calculate_per_group` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        self.name = name
        self.labels = labels
        self.unique_labels = tuple(sorted(set(labels)))

        self.label_map = {lbl: i for i, lbl in enumerate(self.unique_labels)}
        self.label_encodings = np.array(
            [self.label_map[lbl] for lbl in labels], dtype=int
        )

        self.label_indices_map = {
            lbl: np.argwhere(self.label_encodings == self.label_map[lbl]).flatten()
            for lbl in self.unique_labels
        }

    def count(self):
        return {k: len(v) for k, v in self.label_indices_map.items()}

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"UserFeature(name={self.name}, counts={self.count()})"

    def __iter__(self):
        for k, v in self.label_indices_map.items():
            yield k, v



import torch
import numpy as np
from enum import Enum
import scipy.sparse as sp
from scipy.stats import rankdata
from typing import Iterable
from collections import defaultdict

from .type_helpers import (
    as_float,
    assert_supported_type,
    get_top_k,
    get_unique_values,
    generate_non_zero_mask,
    stack,
    std,
    zeros_float,
    zeros_like_float,
)


class MetricEnum(str, Enum):
    DCG = "dcg"
    NDCG = "ndcg"
    Precision = "precision"
    Recall = "recall"
    F_Score = "f_score"
    Hitrate = "hitrate"
    Coverage = "coverage"
    AP = "ap"
    RR = "rr"
    AverageRank = "average_rank"

    def __str__(self):
        return self.value


def _get_relevancy_scores(
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    indices: torch.Tensor | np.ndarray,
):
    if isinstance(targets, torch.Tensor):
        return torch.gather(targets, dim=-1, index=indices)

    elif isinstance(targets, (np.ndarray, sp.csr_array)):
        scores = np.take_along_axis(targets, indices, axis=-1)
        # if there are zeros somewhere in the selected range,
        # take_along_axis returns a sparse array instead of a numpy array
        if isinstance(scores, sp.csr_array):
            return scores.todense()
        return scores
    else:
        assert_supported_type(targets)


def _get_top_k_relevancies(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k=10,
    logits_are_top_indices: bool = False,
    sorted: bool = True,
):
    top_indices = get_top_k(logits, k, logits_are_top_indices, sorted=sorted)
    return _get_relevancy_scores(targets, top_indices)


def _get_n_top_k_relevant(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k=10,
    logits_are_top_indices: bool = False,
    sorted: bool = True,
):
    relevancy_scores = _get_top_k_relevancies(
        logits, targets, k, logits_are_top_indices, sorted=sorted
    )
    return relevancy_scores.sum(-1)


def dcg(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k=10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the Discounted Cumulative Gain (DCG) for items.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    """
    relevancy_scores = _get_top_k_relevancies(
        logits, targets, k, logits_are_top_indices, sorted=True
    )

    if isinstance(relevancy_scores, torch.Tensor):
        discount = 1 / torch.log2(torch.arange(1, k + 1) + 1)
        discount = discount.to(device=logits.device)

    elif isinstance(relevancy_scores, np.ndarray):
        discount = 1 / np.log2(np.arange(1, k + 1) + 1)

    else:
        assert_supported_type(relevancy_scores)

    return as_float(relevancy_scores) @ discount


def ndcg(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the Normalized Discounted Cumulative Gain (nDCG) for items.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    """
    if k <= 0:
        raise ValueError("k is required to be positive!")

    normalization = dcg(targets, targets, k)
    ndcg = dcg(logits, targets, k, logits_are_top_indices) / normalization

    return ndcg


def precision(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the Precision@k (P@k) for items.
    In short, this is the proportion of relevant items in the retrieved items.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    """
    if k <= 0:
        raise ValueError("k is required to be positive!")

    n_relevant_items = _get_n_top_k_relevant(
        logits, targets, k, logits_are_top_indices, sorted=False
    )
    return n_relevant_items / k


def recall(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the Recall@k (R@k) for items.
    In short, this is the proportion of relevant retrieved items of all relevant items.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    """
    n_relevant_items = _get_n_top_k_relevant(
        logits, targets, k, logits_are_top_indices, sorted=False
    )
    n_total_relevant = targets.sum(-1)

    # may happen that there are no relevant true items, cover this possible DivisionByZero case.
    mask = n_total_relevant != 0
    recall = zeros_like_float(n_relevant_items)
    recall[mask] = n_relevant_items[mask] / n_total_relevant[mask]
    return recall


def f_score(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the F-score@k (F@k) for items.
    In short, this is the harmonic mean of precision@k and recall@k.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    """

    p = precision(logits, targets, k, logits_are_top_indices)
    r = recall(logits, targets, k, logits_are_top_indices)

    pr = p + r
    mask = pr != 0
    f_score = zeros_like_float(r)
    f_score[mask] = 2 * ((p * r)[mask] / pr[mask])
    return f_score


def hitrate(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the Hitrate@k (HR@k) for items.
    In short, this is a simple 0/1 metric that considers whether any of the recommended
    items are actually relevant.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    """
    n_relevant_items = _get_n_top_k_relevant(
        logits, targets, k, logits_are_top_indices, sorted=False
    )
    return as_float(n_relevant_items.clip(max=1))


def average_precision(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the mean_average_precision@k (MAP@k) for items.
    In short, it combines precision values at all possible recall levels.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices

    :returns: average precision for each sample of the input
    """
    if k <= 0:
        raise ValueError("k is required to be positive!")

    top_indices = get_top_k(logits, k, logits_are_top_indices, sorted=True)
    n_total_relevant = targets.sum(-1)

    total_precision = zeros_like_float(n_total_relevant)
    for ki in range(1, k + 1):
        # relevance of k'th indices (for -1 see offset in range)
        position_relevance = _get_relevancy_scores(
            targets, top_indices[:, ki - 1 : ki]
        )[:, 0]
        position_precision = precision(
            top_indices, targets, ki, logits_are_top_indices=True
        )
        total_precision += position_precision * position_relevance

    # may happen that there are no relevant true items, cover this possible DivisionByZero case.
    mask = n_total_relevant != 0

    avg_precision = zeros_like_float(n_total_relevant)
    avg_precision[mask] = total_precision[mask] / n_total_relevant[mask]

    return avg_precision


def reciprocal_rank(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the reciprocal rank@k (RR@k) for items.
    In short, it is the inverse rank of the first item that is relevant to the user.
    High values indicate that early items in the recommendations are of interest to the user.
    If there is no relevant item in the top-k recommendations, the reciprocal rank is 0

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices

    :returns: reciprocal rank for each sample of the input
    """
    if k <= 0:
        raise ValueError("k is required to be positive!")

    relevancy_scores = _get_top_k_relevancies(
        logits, targets, k, logits_are_top_indices, sorted=True
    )

    # earliest 'hits' in the recommendation list
    # about determinism, from https://pytorch.org/docs/stable/generated/torch.max.html#torch.max:
    # >>> If there are multiple maximal values in a reduced row
    # >>> then the indices of the first maximal value are returned.
    if isinstance(relevancy_scores, torch.Tensor):
        max_result = torch.max(relevancy_scores, -1)
        max_indices = max_result.indices
        max_values = max_result.values

    elif isinstance(relevancy_scores, np.ndarray | sp.csr_array):
        max_indices = np.argmax(relevancy_scores, -1, keepdims=True)
        max_values = np.take_along_axis(
            relevancy_scores, max_indices, axis=-1
        ).flatten()
        max_indices = max_indices.flatten()
    else:
        assert_supported_type(relevancy_scores)

    # mask to indicate which 'hits' are actually true
    # (if there are no hits at all for some items)
    mask = max_values > 0

    # by default, assume reciprocal rank of 0 for all users,
    # which is the case if there is no match in the recommendations,
    # i.e., if lim k->inf, 1/k->0
    rr = zeros_like_float(max_values)

    denominator = max_indices[mask] + 1
    if isinstance(denominator, torch.Tensor):
        # pytorch is more strict with matching types, so we'll handle it specially
        denominator = denominator.type(rr.dtype)

    # +1 because indices are zero-based, while k is one-based
    rr[mask] = 1.0 / denominator
    return rr


def average_rank(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    targets: torch.Tensor | np.ndarray | sp.csr_array = None,
    item_ranks: torch.Tensor | np.ndarray | sp.csr_array = None,
    k=10,
    logits_are_top_indices: bool = False,
):
    """
    Computes the AverageRank@k (AR@k) for items, which, as the name says, is the
    average rank of the top k items. Ranks are either computed from the given targets,
    where increasing item ranks indicate decreasing popularity, or from the given
    item ranks.

    :param logits: prediction matrix about item relevance
    :param targets: 0/1 matrix encoding true item relevance, same shape as logits
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    :param item_ranks: 1D item ranks to use for computing the average rank

    :returns: average rank for each sample of the input
    """
    if targets is None and item_ranks is None:
        raise ValueError("Either 'targets' or 'item_ranks' must be given.")

    if item_ranks is None:
        score_per_item = targets.sum(0)
        is_tensor_targets = isinstance(targets, torch.Tensor)

        if is_tensor_targets:
            # need to cast to and from torch tensors, as the rank computation
            # is not supported for them
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rankdata.html
            score_per_item = score_per_item.detach().cpu().numpy()

        # get ranks from 1 to n_items, where rank 1 has highest score
        # use "dense" to achieve the same rank for items with same values, but prevents
        # 'jumps' in ranks for the following items.
        # E.g., counts [4,4,3,5,1] should lead to [3,3,2,4,1]
        item_ranks = rankdata(score_per_item, method="dense")
        # flip ranks so that rank 1 is always most popular, rather than least popular
        # as this is better for understanding values across different numbers of items
        item_ranks = max(item_ranks) + 1 - item_ranks

        if is_tensor_targets:
            item_ranks = torch.tensor(item_ranks, device=targets.device)

    top_indices = get_top_k(logits, k, logits_are_top_indices, sorted=False)

    # gather item ranks and average them for each user individually (gather requires matching shapes)
    # this does drop the gradient, but shouldn't be relevant for evaluation metrics
    # another solution would be to gather on item_ranks.repeat((batch_size, 1)), which
    # would allocate more memory
    individual_results = [
        as_float(_get_relevancy_scores(item_ranks, ti)).mean(-1) for ti in top_indices
    ]
    individual_results = stack(individual_results)
    return individual_results


def coverage(
    logits: torch.Tensor | np.ndarray | sp.csr_array,
    k: int = 10,
    logits_are_top_indices: bool = False,
    n_items: int = None,
):
    """
    Computes the Coverage@k (Cov@k) for items.
    In short, this is the proportion of all items that are recommended to the users.

    :param logits: prediction matrix about item relevance
    :param k: top k items to consider
    :param logits_are_top_indices: whether logits are already top-k sorted indices
    :param n_items: if logits are top indices, n_items are requried for coverage computation
    """
    if logits_are_top_indices and n_items is None:
        raise ValueError("'n_items' required when using top indices as logits.")

    if not logits_are_top_indices:
        n_items = logits.shape[-1]

    top_indices = get_top_k(
        logits, k, logits_are_top_indices=logits_are_top_indices, sorted=False
    )
    unique_values = get_unique_values(top_indices[:, :k])
    n_unique_recommended_items = unique_values.shape[0]
    return n_unique_recommended_items / n_items


_metric_fn_map_user_accuracy = {
    MetricEnum.DCG: dcg,
    MetricEnum.NDCG: ndcg,
    MetricEnum.Recall: recall,
    MetricEnum.Precision: precision,
    MetricEnum.Hitrate: hitrate,
    MetricEnum.F_Score: f_score,
    MetricEnum.AP: average_precision,
    MetricEnum.RR: reciprocal_rank,
}

# beyond accuracy metrics may require additional parameters
# logits, k and logits_are_top_indices are implicitly assumed to be necessary
_metric_fn_map_user_beyond_accuracy = {
    MetricEnum.AverageRank: (average_rank, ["targets", "item_ranks"]),
}

_metric_fn_map_global = {MetricEnum.Coverage: coverage}

# List of metrics that are currently supported
supported_metrics = tuple(MetricEnum)
supported_user_accuracy_metrics = tuple(_metric_fn_map_user_accuracy.keys())
supported_user_beyond_accuracy_metrics = tuple(
    _metric_fn_map_user_beyond_accuracy.keys()
)
supported_user_metrics = (
    supported_user_accuracy_metrics + supported_user_beyond_accuracy_metrics
)
supported_global_metrics = tuple(_metric_fn_map_global.keys())


def calculate(
    metrics: str | Iterable[str | MetricEnum],
    logits: torch.Tensor | np.ndarray | sp.csr_array = None,
    targets: torch.Tensor | np.ndarray | sp.csr_array = None,
    k: int | Iterable[int] = 10,
    return_aggregated: bool = True,
    return_per_user: bool = False,
    calculate_std: bool = False,
    flatten_results: bool = False,
    flatten_prefix: str = "",
    flatten_suffix: str = "",
    n_items: int = None,
    best_logit_indices: torch.Tensor | np.ndarray = None,
    return_best_logit_indices: bool = False,
    **kwargs,
):
    """
    Computes the values for a given list of metrics.

    :param metrics:             Metric name or list of metrics to compute. Check out
                                'supported_metrics' for a list of all available metrics.
    :param logits:              prediction matrix about item relevance
    :param targets:             0/1 matrix encoding true item relevance, same shape as logits
    :param k:                   levels of "top-k" items to consider in the calculation
    :param return_aggregated:   whether aggregated metric results should be returned.
    :param return_per_user:     whether the results for individual users should be returned
    :param calculate_std:       whether to calculate the standard deviation for the aggregated results
    :param flatten_results:     whether to flatten the results' dictionary.
    :param flatten_prefix:      prefix to use for flattened results, e.g., to differentiate between
                                different splits
    :param flatten_suffix:      suffix to use for flattened results
    :param n_items:             number of items in dataset (in case only best logit indices are supplied)
    :param best_logit_indices:  previously computed indices of the best logits in sorted order
    :param return_best_logit_indices:   whether to return the indices of the best logits
    :param kwargs:                      additional parameters that are passed to the beyond-accuracy metrics

    :return: a nested dictionary depending of structure {`metric`: {`k`: {`key`: `value`}}},
             where `key` depends on the metric and parameters:
             - "user":   ..., # for user-based metrics if 'return_per_user'
             - "mean":   ..., # for user-based metrics if 'return_aggregated'
             - "std":    ..., # for user-based metrics if 'return_aggregated' and 'calculate_std'
             - "global": ..., # for global metrics
             or a nested dictionary with keys in the format
             - `{flatten_prefix}{metric}@{k}{flatten_suffix}`        # for "mean" and "global" keys
                                                                     # due to their importance
             - `{flatten_prefix}{metric}@{k}_{key}{flatten_suffix}`  # for "std" and "user" keys
    """

    k = (k,) if isinstance(k, int) else k
    max_k = max(k)

    # wrap single metrics in a tuple so that the remaining pipeline works as expected.
    # resulting dictionary output should not change
    metrics = (metrics,) if isinstance(metrics, str) else metrics

    # ensure validity of supplied parameters
    not_supported_metrics = [m for m in metrics if m not in supported_metrics]
    if len(not_supported_metrics) > 0:
        raise ValueError(f"Metrics {not_supported_metrics} are not supported")

    if logits is not None and logits.shape[-1] < max_k:
        raise ValueError(
            f"'k' must not be greater than the number of logits "
            f"({max_k} > {logits.shape[-1]})!"
        )

    if best_logit_indices is not None and best_logit_indices.shape[-1] < max_k:
        raise ValueError(
            f"'k' must not be greater than the number of best indices "
            f"({max_k} > {best_logit_indices.shape[-1]})!"
        )

    if logits is None and (best_logit_indices is None or n_items is None):
        raise ValueError("Either logits or best_logit_indices+n_items must be supplied")

    if best_logit_indices is None and logits.shape != targets.shape:
        raise ValueError(
            f"Logits and targets must be of same shape ({logits.shape} != {targets.shape})"
        )

    if not (return_per_user or return_aggregated):
        raise ValueError(
            f"Specify either 'return_per_user' or 'return_aggregated' to receive results."
        )

    n_items = n_items or logits.shape[-1]
    # to speed up computations, only retrieve highest logit indices once (if not already supplied)
    if best_logit_indices is None:
        best_logit_indices = get_top_k(
            logits, max_k, logits_are_top_indices=False, sorted=True
        )

    # first compute on user-level metrics
    user_metrics = _compute_user_metrics(
        metrics=set(metrics).intersection(set(supported_user_metrics)),
        k=k,
        best_logit_indices=best_logit_indices,
        targets=targets,
        **kwargs,
    )

    # gather all user-based metrics into single dict
    metric_results = defaultdict(lambda: defaultdict(lambda: dict()))
    for m, per_metric_results in user_metrics.items():
        for ki, per_k_results in per_metric_results.items():
            if return_per_user:
                metric_results[m][ki]["user"] = per_k_results

            if return_aggregated:
                metric_results[m][ki]["mean"] = per_k_results.mean().item()
                if calculate_std:
                    metric_results[m][ki]["std"] = std(per_k_results)

    # compute global-based metrics
    global_results = _compute_global_metrics(
        metrics=set(metrics).intersection(set(_metric_fn_map_global)),
        best_logit_indices=best_logit_indices,
        k=k,
        n_items=n_items,
    )
    # ... and add them to results collection
    for m, per_metric_results in global_results.items():
        for ki, k_result in per_metric_results.items():
            metric_results[m][ki]["global"] = k_result

    if flatten_results:
        final_results = dict()
        # iterate over all results and flatten them in single dictionary
        for m, per_metric_results in metric_results.items():
            for ki, per_k_results in per_metric_results.items():
                for key, per_key_results in per_k_results.items():
                    # mean and global metrics are likely the most important metrics,
                    # so to make them look nicer, we drop the key "_{key}" for them
                    # note that both are exclusive, so they will not interfere with
                    # each other
                    if key in ["global", "mean"]:
                        final_key = f"{flatten_prefix}{m}@{ki}{flatten_suffix}"
                    else:
                        final_key = f"{flatten_prefix}{m}@{ki}_{key}{flatten_suffix}"
                    final_results[final_key] = per_key_results
    else:
        # convert defaultdict to normal dict to make return values look cleaner
        final_results = {k: dict(v) for k, v in metric_results.items()}

    if return_best_logit_indices:
        # allow reusing best loggits by returning them,
        # e.g., to call "calculate()" another time
        return dict(final_results), best_logit_indices

    return dict(final_results)


def _compute_user_metrics(
    metrics: Iterable[str | MetricEnum],
    k: Iterable[int],
    best_logit_indices: torch.Tensor | np.ndarray,
    targets: torch.Tensor | np.ndarray | sp.csr_array = None,
    **kwargs,
):
    """
    Computes all user-based metrics

    :param metrics:             the user-based metrics to compute
    :param k:                   top k items to consider
    :param best_logit_indices:  previously computed indices of the best logits in sorted order
    :param targets:             0/1 matrix encoding true item relevance, same shape as logits
    :param kwargs:              additional parameters that are passed to the beyond-accuracy metrics

    :return: a nested dictionary with levels for the 'metrics' and 'k', returning the results
             for individual users, e.g., {
        "ndcg": {
            1:  [0.41, 0.23, 0.19, 0.00, 1.00],
            10: [0.52, 0.37, 0.14, 0.12, 0.79],
        },
    }
    """
    results = {}

    # compute user-based accuracy metrics
    user_metrics_to_compute = set(metrics).intersection(
        set(_metric_fn_map_user_accuracy)
    )
    if len(user_metrics_to_compute):
        results.update(
            _compute_user_accuracy_metrics(
                metrics=user_metrics_to_compute,
                k=k,
                best_logit_indices=best_logit_indices,
                targets=targets,
            )
        )

    # compute user-based beyond accuracy metrics
    user_beyond_metrics_to_compute = set(metrics).intersection(
        set(_metric_fn_map_user_beyond_accuracy)
    )
    if len(user_beyond_metrics_to_compute):
        results.update(
            _compute_user_beyond_accuracy_metrics(
                metrics=user_beyond_metrics_to_compute,
                k=k,
                best_logit_indices=best_logit_indices,
                targets=targets,
                **kwargs,
            )
        )

    return results


def _compute_user_accuracy_metrics(
    metrics: Iterable[str | MetricEnum],
    k: Iterable[int],
    best_logit_indices: torch.Tensor | np.ndarray,
    targets: torch.Tensor | np.ndarray | sp.csr_array,
):
    """
    Computes the given user-based accuracy metrics

    :param metrics:             the metrics to compute
    :param k:                   top k items to consider
    :param best_logit_indices:  previously computed indices of the best logits in sorted order
    :param targets:             0/1 matrix encoding true item relevance, same shape as logits

    :return: a nested dictionary with levels for the 'metrics' and 'k', returning the results
             for individual users, e.g., {
        "ndcg": {
            1:  [0.41, 0.23, 0.19, 0.00, 1.00],
            10: [0.52, 0.37, 0.14, 0.12, 0.79],
        },
    }
    """
    results = {}

    if targets is None:
        raise ValueError(f"'targets' is required to calculate '{metric}'!")

    # do not compute metrics for users where we do not have any
    # underlying ground truth interactions
    mask = generate_non_zero_mask(targets)

    results = defaultdict(lambda: dict())
    for ki in k:
        for metric in metrics:
            # compute metrics only for users with targets
            metric_result = zeros_float(targets.shape[0], targets)
            metric_result[mask] = _metric_fn_map_user_accuracy[metric](
                logits=best_logit_indices[mask],
                targets=targets[mask],
                k=ki,
                logits_are_top_indices=True,
            )
            results[str(metric)][ki] = metric_result

    return dict(results)


def _compute_user_beyond_accuracy_metrics(
    metrics: Iterable[str | MetricEnum],
    k: Iterable[int],
    best_logit_indices: torch.Tensor | np.ndarray,
    **kwargs,
):
    """
    Computes the given user-based beyond-accuracy metrics

    :param metrics:             the metrics to compute
    :param k:                   top k items to consider
    :param best_logit_indices:  previously computed indices of the best logits in sorted order
    :param targets:             0/1 matrix encoding true item relevance, same shape as logits
    :param kwargs:              additional parameters that are passed to the beyond-accuracy metrics

    :return: a nested dictionary with levels for the 'metrics' and 'k', returning the results
             for individual users, e.g., {
        "average_rank": {
            1:  [0.41, 0.23, 0.19, 0.00, 1.00],
            10: [0.52, 0.37, 0.14, 0.12, 0.79],
        },
    }
    """
    results = defaultdict(lambda: dict())
    for ki in k:
        for metric in metrics:
            fn, required_params = _metric_fn_map_user_beyond_accuracy[metric]
            results[str(metric)][ki] = fn(
                logits=best_logit_indices,
                k=ki,
                logits_are_top_indices=True,
                **{p: kwargs.get(p) for p in required_params},
            )
    return dict(results)


def _compute_global_metrics(
    metrics: Iterable[str | MetricEnum],
    k: Iterable[int],
    best_logit_indices: torch.Tensor | np.ndarray,
    n_items: int,
):
    """
    Computes the given global metrics

    :param metrics:             the metrics to compute
    :param k:                   top k items to consider
    :param best_logit_indices:  previously computed indices of the best logits in sorted order
    :param n_items:             number of items in dataset (in case only best logit indices are supplied)

    :return: a nested dictionary with levels for the 'metrics' and 'k', returning the
             global results, e.g., {
                 "coverage": {
                    1:  0.01,
                    10: 0.21,
                 }
             }}
    """
    results = defaultdict(lambda: dict())
    for ki in k:
        for metric in metrics:
            results[str(metric)][ki] = _metric_fn_map_global[metric](
                logits=best_logit_indices,
                k=ki,
                n_items=n_items,
                logits_are_top_indices=True,
            )
    return dict(results)



import math
import torch
import unittest
import numpy as np
import scipy.sparse as sp
from collections import defaultdict

from rmet.metrics import (
    average_rank,
    coverage,
    precision,
    recall,
    dcg,
    ndcg,
    hitrate,
    f_score,
    average_precision,
    reciprocal_rank,
    calculate,
    supported_metrics,
)

from rmet.batch_evaluator import BatchEvaluator, EvaluatorResults
from rmet.type_helpers import std


class TestRecommenderMetrics(unittest.TestCase):

    def setUp(self):
        # top-k: [0, 1, 3], [3, 4, 2]
        self.logits = [[0.9, 0.8, 0, 0.4, 0], [0.0, 0.0, 0.3, 0.9, 0.8]]
        # relevant: [0, 2, 3], [3, 4]
        self.targets = [[1, 0, 1, 1, 0], [0, 0, 0, 1, 1]]

        self.supported_types = {
            "torch": {
                "cast": torch.tensor,
                "cast-result": torch.tensor,
                "all_close": torch.allclose,
                "zeros_like": torch.zeros_like,
            },
            "np": {
                "cast": np.array,
                "cast-result": np.array,
                "all_close": np.allclose,
                "zeros_like": np.zeros_like,
            },
            "sparse-csr": {
                "cast": sp.csr_array,
                "cast-result": np.array,
                "all_close": np.allclose,
                "zeros_like": np.zeros_like,
            },
        }

        self.k = 3
        self.best_logit_indices = [[0, 1, 3], [3, 4, 2]]

        self.precision1 = 2 / 3
        self.precision2 = 2 / 3
        self.precision = [self.precision1, self.precision2]

        self.recall1 = 2 / 3
        self.recall2 = 1.0
        self.recall = [self.recall1, self.recall2]

        self.dcg1 = 1 + 0 + 1 / math.log2(4)
        self.dcg2 = 1 + 1 / math.log2(3)
        self.dcg = [self.dcg1, self.dcg2]

        self.ideal_dcg1 = 1 + 1 / math.log2(3) + 1 / math.log2(4)
        self.ideal_dcg2 = 1 + 1 / math.log2(3)
        self.ideal_dcg = [self.ideal_dcg1, self.ideal_dcg2]

        self.ndcg1 = self.dcg1 / self.ideal_dcg1
        self.ndcg2 = self.dcg2 / self.ideal_dcg2
        self.ndcg = [self.ndcg1, self.ndcg2]

        self.hitrate1 = 1.0
        self.hitrate2 = 1.0
        self.hitrate = [self.hitrate1, self.hitrate2]

        self.f1_user1 = (
            2 * self.precision1 * self.recall1 / (self.precision1 + self.recall1)
        )
        self.f1_user2 = (
            2 * self.precision2 * self.recall2 / (self.precision2 + self.recall2)
        )
        self.f1_user = [self.f1_user1, self.f1_user2]

        self.ap1 = (1 / 1 + 2 / 3) / 3
        self.ap2 = (1 / 1 + 2 / 2) / 2
        self.ap = [self.ap1, self.ap2]

        # first relevant at rank 1 for both users
        self.rr1 = 1.0
        self.rr2 = 1.0
        self.rr = [self.rr1, self.rr2]

        self.coverage = 5 / 5

        self.item_ranks = [5, 3, 2, 4, 1]
        self.ar1 = (5 + 3 + 4) / 3
        self.ar2 = (2 + 4 + 1) / 3
        self.ar = [self.ar1, self.ar2]

        # computed item ranks should be [2, 3, 2, 1, 2],
        # as interaction counts are [1, 0, 1, 2, 1]
        # (same values are assigned same ranks, without gaps between ranks)
        self.ar1_from_targets = (2 + 3 + 1) / 3
        self.ar2_from_targets = (1 + 2 + 2) / 3
        self.ar_from_targets = [self.ar1_from_targets, self.ar2_from_targets]

        self.user_computation_results = {
            "dcg": self.dcg,
            "ndcg": self.ndcg,
            "recall": self.recall,
            "precision": self.precision,
            "hitrate": self.hitrate,
            "f_score": self.f1_user,
            "ap": self.ap,
            "rr": self.rr,
            "average_rank": self.ar,
        }
        self.metrics = list(self.user_computation_results.keys()) + ["coverage"]

    def _assert_dictionary_equality(self, d1, d2, is_close_fn):
        # check matching keys
        self.assertEqual(type(d1), type(d2))
        self.assertSetEqual(set(d1.keys()), set(d2.keys()))
        for k in d1:
            # dict might be nested
            if isinstance(d1[k], dict):
                self._assert_dictionary_equality(d1[k], d2[k], is_close_fn)
            else:
                # check matching results
                self.assertEqual(type(d1[k]), type(d2[k]))
                if isinstance(d1[k], (np.ndarray, torch.Tensor)):
                    self.assertTrue(is_close_fn(d1[k], d2[k]))
                else:
                    self.assertAlmostEqual(d1[k], d2[k], delta=1e-4)

    def test_precision(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.precision)
            result = precision(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_recall(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.recall)
            result = recall(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_dcg(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.dcg)
            result = dcg(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_ndcg(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.ndcg)
            result = ndcg(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_hitrate(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.hitrate)
            result = hitrate(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_f_score(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.f1_user)
            result = f_score(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_average_precision(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.ap)
            result = average_precision(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_reciprocal_rank(self):
        for _, fn_lookup in self.supported_types.items():
            expected = fn_lookup["cast-result"](self.rr)
            result = reciprocal_rank(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_average_rank(self):
        for _, fn_lookup in self.supported_types.items():
            # compute WITHOUT item ranks
            expected = fn_lookup["cast-result"](self.ar_from_targets)
            result = average_rank(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

            # compute WITH item ranks
            expected = fn_lookup["cast-result"](self.ar)
            result = average_rank(
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
                item_ranks=fn_lookup["cast"](self.item_ranks),
            )
            self.assertTrue(fn_lookup["all_close"](result, expected, atol=1e-4))

    def test_all_zero_targets(self):
        for _, fn_lookup in self.supported_types.items():
            logits = fn_lookup["cast"](self.logits)
            targets = fn_lookup["zeros_like"](fn_lookup["cast-result"](self.targets))
            k = self.k

            zero = fn_lookup["cast-result"]([0.0, 0.0])
            self.assertTrue(fn_lookup["all_close"](precision(logits, targets, k), zero))
            self.assertTrue(fn_lookup["all_close"](recall(logits, targets, k), zero))
            self.assertTrue(fn_lookup["all_close"](hitrate(logits, targets, k), zero))
            self.assertTrue(fn_lookup["all_close"](f_score(logits, targets, k), zero))
            self.assertTrue(
                fn_lookup["all_close"](average_precision(logits, targets, k), zero)
            )
            self.assertTrue(
                fn_lookup["all_close"](reciprocal_rank(logits, targets, k), zero)
            )

    def test_coverage(self):
        for _, fn_lookup in self.supported_types.items():
            logits = fn_lookup["cast"](self.logits)

            self.assertAlmostEqual(coverage(logits, self.k), self.coverage, delta=1e-4)
            self.assertAlmostEqual(coverage(logits, 2), 4 / 5, delta=1e-4)

    def test_compute_nested(self):
        for _, fn_lookup in self.supported_types.items():
            result = calculate(
                metrics=self.metrics,
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
                return_aggregated=True,
                return_per_user=True,
                calculate_std=True,
                flatten_results=False,
                n_items=None,
                best_logit_indices=None,
                return_best_logit_indices=False,
                item_ranks=fn_lookup["cast"](self.item_ranks),
            )

            expected = defaultdict(lambda: defaultdict(lambda: dict()))
            for m, r in self.user_computation_results.items():
                expected[m][self.k]["user"] = fn_lookup["cast-result"](r)
                expected[m][self.k]["mean"] = np.mean(r).item()
                # ddof to adjust to degrees of freedom = 1 for std computation in PyTorch
                expected[m][self.k]["std"] = np.std(r, ddof=1).item()
            expected.update({"coverage": {self.k: {"global": self.coverage}}})
            # transform to normal dictionary
            expected = {k: dict(v) for k, v in expected.items()}

            self._assert_dictionary_equality(result, expected, fn_lookup["all_close"])

    def test_compute_flattened(self):

        prefix = "<my-prefix>/"
        suffix = "/<my-suffix>"

        for _, fn_lookup in self.supported_types.items():
            result = calculate(
                metrics=self.metrics,
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
                return_aggregated=True,
                return_per_user=True,
                calculate_std=True,
                flatten_results=True,
                flatten_prefix=prefix,
                flatten_suffix=suffix,
                n_items=None,
                best_logit_indices=None,
                return_best_logit_indices=False,
                item_ranks=fn_lookup["cast"](self.item_ranks),
            )

            expected = dict()
            for m, k in self.user_computation_results.items():
                metric_name = f"{prefix}{m}@{self.k}"
                expected[f"{metric_name}_user" + suffix] = fn_lookup["cast-result"](k)
                expected[metric_name + suffix] = np.mean(k).item()
                # ddof to adjust to degrees of freedom = 1 for std computation in PyTorch
                expected[f"{metric_name}_std" + suffix] = np.std(k, ddof=1).item()
            expected.update({f"{prefix}coverage@{self.k}{suffix}": self.coverage})

            self._assert_dictionary_equality(result, expected, fn_lookup["all_close"])

    def test_best_indices(self):
        for _, fn_lookup in self.supported_types.items():
            result, best_indices = calculate(
                metrics=self.metrics,
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
                return_best_logit_indices=True,
                return_per_user=True,
                flatten_results=True,
                item_ranks=fn_lookup["cast"](self.item_ranks),
            )

            self.assertTrue(
                fn_lookup["all_close"](
                    best_indices, fn_lookup["cast-result"](self.best_logit_indices)
                )
            )

            result_on_best_logits = calculate(
                metrics=self.metrics,
                logits=fn_lookup["cast"](self.logits),
                targets=fn_lookup["cast"](self.targets),
                k=self.k,
                best_logit_indices=best_indices,
                return_best_logit_indices=False,
                return_per_user=True,
                flatten_results=True,
                item_ranks=fn_lookup["cast"](self.item_ranks),
            )

            self._assert_dictionary_equality(
                result, result_on_best_logits, fn_lookup["all_close"]
            )

    def test_single_metric_vs_list(self):
        # test string metrics
        for m in supported_metrics:
            for _, fn_lookup in self.supported_types.items():
                r_single_metric = calculate(
                    metrics=m,
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                )

                r_metrics_list = calculate(
                    metrics=m,
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                )

                self._assert_dictionary_equality(
                    r_single_metric, r_metrics_list, fn_lookup["all_close"]
                )

    def test_accepted_metrics(self):
        # test enum metrics
        for m in supported_metrics:
            for _, fn_lookup in self.supported_types.items():
                calculate(
                    metrics=[m],
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                )

        # test string metrics
        for m in supported_metrics:
            for _, fn_lookup in self.supported_types.items():
                calculate(
                    metrics=[str(m)],
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                )

        # test mix between string and Enum metrics
        for i, m in enumerate(supported_metrics):
            for _, fn_lookup in self.supported_types.items():
                m = m if i % 2 == 0 else str(m)
                print(m)
                calculate(
                    metrics=[str(m)],
                    logits=fn_lookup["cast"](self.logits),
                    targets=fn_lookup["cast"](self.targets),
                    k=self.k,
                )

        # test some random metrics that should fail
        not_supported_metrics = ["foo", "bar", "foobar", "barfoo"]
        for m in not_supported_metrics:
            for _, fn_lookup in self.supported_types.items():
                with self.assertRaises(ValueError):
                    calculate(
                        metrics=[m],
                        logits=fn_lookup["cast"](self.logits),
                        targets=fn_lookup["cast"](self.targets),
                        k=self.k,
                    )

    def test_batch_evaluator(self):
        for _, fn_lookup in self.supported_types.items():
            batch_evaluator = BatchEvaluator(
                metrics=self.metrics,
                top_k=self.k,
                calculate_std=True,
                n_items=None,
                **{"item_ranks": fn_lookup["cast"](self.item_ranks)},
            )

            for i in range(len(self.logits)):
                batch_evaluator.eval_batch(
                    user_indices=fn_lookup["cast-result"]([i]),
                    # slice to maintain batch
                    logits=fn_lookup["cast"](self.logits[i : i + 1]),
                    targets=fn_lookup["cast"](self.targets[i : i + 1]),
                )
            result = batch_evaluator.get_results()

            expected = EvaluatorResults(
                user_level_metrics={},
                aggregated_metrics={},
                user_indices=None,
                user_top_k=None,
            )

            for m, m_result in self.user_computation_results.items():
                metric_name = f"{m}@{self.k}"
                m_result = fn_lookup["cast-result"](m_result)
                expected.user_level_metrics[metric_name] = m_result
                expected.aggregated_metrics[metric_name] = m_result.mean().item()
                # ddof to adjust to degrees of freedom = 1 for std computation in PyTorch,
                # as we want to have the same results between NumPy and PyTorch
                expected.aggregated_metrics[metric_name + "_std"] = std(m_result)
            expected.aggregated_metrics.update({f"coverage@{self.k}": self.coverage})

            self._assert_dictionary_equality(
                result.user_level_metrics, expected.user_level_metrics, np.allclose
            )
            self._assert_dictionary_equality(
                result.aggregated_metrics, expected.aggregated_metrics, np.allclose
            )


if __name__ == "__main__":
    unittest.main()
